"""Tests for KLDivergence Power Spectrum."""

from copy import deepcopy
from itertools import pairwise
from typing import Tuple

import numpy as np
import pytest
import torch as th

from scripts.wpkl.wpkl import compute_packets, wavelet_power_divergence
from tests.test_wavelet_frechet_distance import (
    default_params,
    get_images,
    make_dataloader,
)

th.set_default_dtype(th.float64)
th.use_deterministic_algorithms(True)


def _calc_klwd(target_images: th.Tensor, output_images: th.Tensor) -> float:
    """Compute KLDivergence Power spectrum.

    Args:
        target_images (th.Tensor): Ground Truth images
        output_images (th.Tensor): Generated images

    Returns:
        float: WPKL value.
    """
    default_params["dataloader"] = make_dataloader(target_images)
    target_packets = compute_packets(**default_params)
    default_params["dataloader"] = make_dataloader(output_images)
    output_packets = compute_packets(**default_params)

    klwd = wavelet_power_divergence(target_packets, output_packets)
    return klwd


def test_same_input():
    """WPKL-test for same input."""
    target_images = get_images()
    output_images = deepcopy(target_images)
    klwd = _calc_klwd(target_images, output_images)
    assert np.allclose(klwd, 0.0)


def test_shuffle_input():
    """WPKL-test for shuffled input."""
    target_images = get_images()
    output_images = deepcopy(target_images)
    permutation = th.randperm((len(target_images)))
    shuffled_images = output_images[permutation, :, :, :]
    assert not th.allclose(shuffled_images, output_images)

    shuffled_klwd = _calc_klwd(target_images, shuffled_images)
    unshuffled_klwd = _calc_klwd(target_images, output_images)
    assert np.allclose(shuffled_klwd, unshuffled_klwd)
    assert np.allclose(shuffled_klwd, 0.0)
    assert np.allclose(unshuffled_klwd, 0.0)


@pytest.mark.slow
@pytest.mark.parametrize("wavelet", ["sym5", "db5", "Haar"])
@pytest.mark.parametrize("level", [2, 3])
def test_various_wavelets(wavelet: str, level: int):
    """WPKL-test for various wavelets and transformation levels.

    Args:
        wavelet (str): Type of wavelet. Accepted values [sym5, db5, Haar]
        level (int): Packet transformation level. Accepted values [2, 3].
    """
    target_images = get_images()
    output_images = deepcopy(target_images)

    default_params["wavelet"] = wavelet
    default_params["max_level"] = level

    fwd = _calc_klwd(target_images, output_images)
    assert np.allclose(fwd, 0.0, atol=1e-3)


@pytest.mark.slow
@pytest.mark.parametrize("img_size_level", [(32, 1), (64, 2)])
def test_various_image_sizes(img_size_level: Tuple[int, int]):
    """Test various image sizes and corresponding transfromation levels.

    Args:
        img_size_level (Tuple[int, int]): Tuple containing img size and transfromation level.
    """
    size, level = img_size_level

    default_params["max_level"] = level

    target_images = get_images(size)
    output_images = deepcopy(target_images)
    assert output_images.shape == (8, 3, size, size)

    fwd = _calc_klwd(target_images, output_images)
    assert np.allclose(fwd, 0.0, atol=1e-3)


@pytest.mark.slow
def test_checkerboard_power_kldiv():
    """WPKL-Synthetic frequency test using various checkerboard sizes."""
    import scipy.ndimage

    range_max = 100
    # tile_size = 10 # determines tile size
    images = []

    for tile_size in [2, 5, 10]:
        # generate random grid

        grid = np.meshgrid(np.arange(0, range_max), np.arange(0, range_max))
        # grid = grid[0] + grid[1]

        tile_fun = lambda cord, ts: (cord % ts) < (ts // 2)  # type: ignore # noqa: E731

        x_tile = tile_fun(grid[0], tile_size)
        y_tile = tile_fun(grid[1], tile_size)
        grid = 1 - (x_tile + y_tile)

        grid = grid.astype(np.float32)

        # generate 10 images via a random fft-shift
        img_list = []
        for _ in range(1000):
            input_ = np.fft.fft2(grid)
            shift = np.random.uniform(-1, 1) * tile_size // 2
            image = scipy.ndimage.fourier_shift(input_, shift=shift)
            image = np.fft.ifft2(image)
            img_list.append(image.real)
        # B, H, W
        img_array = np.stack(img_list, axis=0)
        # B, C, H, W
        img_array = np.expand_dims(img_array, 1)
        # img_array = np.stack([img_array]*3, 1) # TODO!
        images.append(img_array)

    # a list for the wavelet frecet distances.
    wfd_list = []
    default_params["max_level"] = 3
    reference = images[0]
    for cimgs in images[1:]:
        kwd = _calc_klwd(
            target_images=th.from_numpy(reference), output_images=th.from_numpy(cimgs)
        )
        wfd_list.append(kwd)
    assert all(a < b for a, b in pairwise(wfd_list))



"""Frechet Wavelet Distance computation."""

import os
import pathlib
from typing import List, Tuple

import numpy as np
import torch as th
import torchvision.transforms as tv
from tqdm import tqdm

from .freq_math import calculate_frechet_distance, forward_wavelet_packet_transform
from .utils import ImagePathDataset, _parse_args

th.set_default_dtype(th.float64)


IMAGE_EXTS = {"jpg", "jpeg", "png"}
NUM_PROCESSES = None


def compute_packet_statistics(
    dataloader: th.utils.data.DataLoader, wavelet: str, max_level: int, log_scale: bool
) -> Tuple[np.ndarray, ...]:
    """Compute wavelet packet transform across batches.

    Args:
        dataloader (th.utils.data.DataLoader): Torch dataloader.
        wavelet (str): Choice of wavelet.
        max_level (int): Wavelet decomposition level.
        log_scale (bool): Apply log scale.

    Returns:
        Tuple[np.ndarray, ...]: Mean and sigma for each packet.
    """
    packets = []
    device = th.device("cuda:0") if th.cuda.is_available() else th.device("cpu")
    for img_batch in tqdm(dataloader):
        if isinstance(img_batch, list):
            img_batch = img_batch[0]
        img_batch = img_batch.to(device)
        packets.append(
            forward_wavelet_packet_transform(
                img_batch, wavelet, max_level, log_scale
            ).cpu()
        )
    packet_tensor = th.cat(packets, dim=0)
    packet_tensor = th.permute(packet_tensor, (1, 0, 2, 3, 4))
    P, BS, C, H, W = packet_tensor.shape
    packet_tensor = th.reshape(packet_tensor, (P, BS, C * H * W))
    print("Computing mean and std for each packet.")
    mu = th.mean(packet_tensor, dim=1).numpy()

    def gpu_cov(tensor_):
        return th.cov(tensor_.T).cpu()

    sigma = th.stack(
        [gpu_cov(packet_tensor[p, :, :].to(device)) for p in range(P)], dim=0
    ).numpy()
    return mu, sigma


def calculate_path_statistics(
    path: str, wavelet: str, max_level: int, log_scale: bool, batch_size: int
) -> Tuple[np.ndarray, ...]:
    """Compute mean and sigma for given path.

    Args:
        path (str): npz path or image directory.
        wavelet (str): Choice of wavelet.
        max_level (int): Decomposition level.
        log_scale (bool): Apply log scale.
        batch_size (int): Batch size for packet decomposition.

    Raises:
        ValueError: Error if mu and sigma cannot be calculated.

    Returns:
        Tuple[np.ndarray, ...]: Tuple containing mean and sigma for each packet.
    """
    mu, sigma = None, None
    if path.endswith(".npz") or path.endswith(".npy"):
        with np.load(path) as fp:
            mu = fp["mu"][:]
            sigma = fp["sigma"][:]
    else:
        posfix_path = pathlib.Path(path)
        img_names = sorted(
            [name for ext in IMAGE_EXTS for name in posfix_path.glob(f"*.{ext}")]
        )
        dataloader = th.utils.data.DataLoader(
            ImagePathDataset(img_names, transforms=tv.ToTensor()),
            batch_size=batch_size,
            shuffle=False,
            drop_last=False,
            num_workers=NUM_PROCESSES,
        )
        mu, sigma = compute_packet_statistics(
            dataloader=dataloader,
            wavelet=wavelet,
            max_level=max_level,
            log_scale=log_scale,
        )

    if (mu is None) or (sigma is None):
        raise ValueError(f"The file path: {path} is empty/doesn't have statistics.")
    return mu, sigma


def _compute_avg_frechet_distance(mu1, mu2, sigma1, sigma2):
    """Compute avg frechet distance over packets."""
    frechet_distances = []
    for packet_no in tqdm(range(len(mu1))):
        fd = calculate_frechet_distance(
            mu1=mu1[packet_no, :],
            mu2=mu2[packet_no, :],
            sigma1=sigma1[packet_no, :, :],
            sigma2=sigma2[packet_no, :, :],
        )
        frechet_distances.append(fd)
    return np.mean(frechet_distances)


def compute_fwd(
    paths: List[str], wavelet: str, max_level: int, log_scale: bool, batch_size: int
) -> float:
    """Compute Frechet Wavelet Distance.

    Args:
        paths (List[str]): List containing path of source and generated images.
        wavelet (str): Choice of wavelet.
        max_level (int): Decomposition level.
        log_scale (bool): Apply log scale.
        batch_size (int): Batch size for packet decomposition.

    Raises:
        RuntimeError: Error if path doesn't exist.

    Returns:
        float: Frechet Wavelet Distance.
    """
    for path in paths:
        if not os.path.exists(path):
            raise RuntimeError(f"Invalid path: {path}")

    print(f"Computing stats for path: {paths[0]}")
    mu_1, sigma_1 = calculate_path_statistics(
        paths[0], wavelet, max_level, log_scale, batch_size
    )
    print(f"Computing stats for path: {paths[1]}")
    mu_2, sigma_2 = calculate_path_statistics(
        paths[1], wavelet, max_level, log_scale, batch_size
    )

    print("Computing Frechet distances for each packet.")
    return _compute_avg_frechet_distance(mu_1, mu_2, sigma_1, sigma_2)


def _save_packets(
    paths: List[str], wavelet: str, max_level: int, log_scale: bool, batch_size: int
) -> None:
    """Save packets.

    Args:
        paths (List[str]): List of paths containing input and output files.
        wavelet (str): Choice of wavelet.
        max_level (int): Decomposition level.
        log_scale (bool): Apply log scale.
        batch_size (int): Batch size for packet decomposition.

    Raises:
        RuntimeError: Error if input path is invalid.
        RuntimeError: Error if the output file already exists.
    """
    if not os.path.exists(paths[0]):
        raise RuntimeError(f"Invalid path: {paths[0]}")

    if os.path.exists(paths[1]):
        raise RuntimeError(f"Stats file already exists at the given path: {paths[1]}")

    print(f"Computing stats for path: {paths[0]}")
    mu_1, sigma_1 = calculate_path_statistics(
        paths[0], wavelet, max_level, log_scale, batch_size
    )
    np.savez_compressed(paths[1], mu=mu_1, sigma=sigma_1)


def main():
    """Compute FWD given paths."""
    global NUM_PROCESSES, IMAGE_EXTS

    th.manual_seed(0)
    args = _parse_args()
    print(args)
    if args.num_processes is None:
        try:
            num_cpus = len(os.sched_getaffinity(0))
        except AttributeError:
            num_cpus = os.cpu_count()
        NUM_PROCESSES = min(num_cpus, 16) if num_cpus is not None else 0
    else:
        NUM_PROCESSES = args.num_processes
    print(f"Num work: {NUM_PROCESSES}")
    if args.deterministic:
        th.use_deterministic_algorithms(True)
    if args.save_packets:
        _save_packets(
            args.path, args.wavelet, args.max_level, args.log_scale, args.batch_size
        )
        return

    fwd = compute_fwd(
        args.path, args.wavelet, args.max_level, args.log_scale, args.batch_size
    )
    print(f"FWD: {fwd}")


if __name__ == "__main__":
    main()



"""Test Frechet Inception distance."""

from copy import deepcopy
from typing import Tuple

import numpy as np
import pytest
import torch as th
from torch.nn.functional import adaptive_avg_pool2d

from pytorchfwd.freq_math import calculate_frechet_distance
from scripts.fid.inception import InceptionV3

from .test_wavelet_frechet_distance import get_images


@pytest.mark.slow
def forward_pass(images: th.Tensor) -> Tuple[np.ndarray, ...]:
    """Compute InceptionV3 forward pass.

    Args:
        images (th.Tensor): Image tensor.

    Returns:
        Tuple: Tuple containing mu and sigma of predictions.
    """
    block_idx = InceptionV3.BLOCK_INDEX_BY_DIM[2048]
    model = InceptionV3([block_idx])
    with th.no_grad():
        pred = model(images)[0]
    if pred.size(2) != 1 or pred.size(3) != 1:
        pred = adaptive_avg_pool2d(pred, output_size=(1, 1))
    pred = pred.squeeze(3).squeeze(2).cpu().numpy()
    assert pred.shape == (images.shape[0], 2048)
    mu = np.mean(pred, axis=0)
    sigma = np.cov(pred, rowvar=False)
    return mu, sigma


@pytest.mark.slow
def test_same_input():
    """FID-test same input."""
    tensor_images = get_images(img_size=256)
    output_images = deepcopy(tensor_images)
    mu1, sigma1 = forward_pass(tensor_images)
    mu2, sigma2 = forward_pass(output_images)
    fid = calculate_frechet_distance(mu1, sigma1, mu2, sigma2)
    assert np.allclose(fid, 0.0, atol=1e-3)


@pytest.mark.slow
def test_shuffle_input():
    """FID-test shuffled input."""
    tensor_images = get_images(img_size=256)
    output_images = deepcopy(tensor_images)
    permutation = th.randperm(len(output_images))
    shuffled_images = output_images[permutation, :, :, :]
    assert not th.allclose(shuffled_images, output_images)
    mu_orig, sigma_orig = forward_pass(tensor_images)
    mu_shuff, sigma_shuff = forward_pass(shuffled_images)
    mu_op, sigma_op = forward_pass(output_images)
    original_fid = calculate_frechet_distance(mu_orig, sigma_orig, mu_op, sigma_op)
    shuffled_fid = calculate_frechet_distance(
        mu_orig, sigma_orig, mu_shuff, sigma_shuff
    )
    assert np.allclose(shuffled_fid, original_fid, atol=1e-4)



"""Wavelet utils."""

from itertools import product
from typing import Optional

import numpy as np
import ptwt
import pywt
import torch
from scipy import linalg


def get_freq_order(level: int):
    """Get the frequency order for a given packet decomposition level.

    Adapted from:
    https://github.com/PyWavelets/pywt/blob/master/pywt/_wavelet_packets.py
    The code elements denote the filter application order. The filters
    are named following the pywt convention as:
    a - LL, low-low coefficients
    h - LH, low-high coefficients
    v - HL, high-low coefficients
    d - HH, high-high coefficients
    """
    wp_natural_path = list(product(["a", "h", "v", "d"], repeat=level))

    def _get_graycode_order(level, x="a", y="d"):
        graycode_order = [x, y]
        for _ in range(level - 1):
            graycode_order = [x + path for path in graycode_order] + [
                y + path for path in graycode_order[::-1]
            ]
        return graycode_order

    def _expand_2d_path(path):
        expanded_paths = {"d": "hh", "h": "hl", "v": "lh", "a": "ll"}
        return (
            "".join([expanded_paths[p][0] for p in path]),
            "".join([expanded_paths[p][1] for p in path]),
        )

    nodes: dict = {}
    for (row_path, col_path), node in [
        (_expand_2d_path(node), node) for node in wp_natural_path
    ]:
        nodes.setdefault(row_path, {})[col_path] = node
    graycode_order = _get_graycode_order(level, x="l", y="h")
    nodes_list: list = [nodes[path] for path in graycode_order if path in nodes]
    wp_frequency_path = []
    for row in nodes_list:
        wp_frequency_path.append([row[path] for path in graycode_order if path in row])
    return wp_frequency_path, wp_natural_path


def forward_wavelet_packet_transform(
    tensor: torch.Tensor,
    wavelet: str,
    max_level: int,
    log_scale: bool,
) -> torch.Tensor:
    """Compute wavelet packet transform.

    Args:
        tensor (torch.Tensor): Input torch tensor
        wavelet (str): Choice of wavelet
        max_level (int): Level of decomposition
        log_scale (bool): Log scale boolean

    Returns:
        torch.Tensor: Packets
    """
    # ideally the output dtype should depend in the input.
    # tensor = tensor.type(torch.FloatTensor)
    packets = ptwt.WaveletPacket2D(tensor, pywt.Wavelet(wavelet), maxlevel=max_level)
    packet_list = [packets[node] for node in packets.get_natural_order(max_level)]

    # for node in packets.get_natural_order(max_level):
    # packet = torch.squeeze(packets[node], dim=1)
    # packet_list.append(packets[node])
    wp_pt_rs = torch.stack(packet_list, dim=1)
    if log_scale:
        wp_pt_rs = torch.log(torch.abs(wp_pt_rs) + 1e-6)

    return wp_pt_rs


def generate_frequency_packet_image(packet_array: np.ndarray, degree: int):
    """Create a ready-to-polt image with frequency-order packages.

       Given a packet array in natural order, creat an image which is
       ready to plot in frequency order.
    Args:
        packet_array (np.ndarray): [packet_no, packet_height, packet_width]
            in natural order.
        degree (int): The degree of the packet decomposition.
    Returns:
        [np.ndarray]: The image of shape [original_height, original_width]
    """
    wp_freq_path, wp_natural_path = get_freq_order(degree)

    image = []
    # go through the rows.
    for row_paths in wp_freq_path:
        row = []
        for row_path in row_paths:
            index = wp_natural_path.index(row_path)
            packet = packet_array[:, index]
            row.append(packet)
        image.append(np.concatenate(row, -1))
    return np.concatenate(image, 2)


def compute_kl_divergence(
    output: torch.Tensor, target: torch.Tensor, eps: Optional[float] = 1e-30
) -> torch.Tensor:
    """Compute KL Divergence.

    Args:
        output (torch.Tensor): Output Images
        target (torch.Tensor): Target Images
        eps (float, optional): Epsilon. Defaults to 1e-12.

    Returns:
        torch.Tensor: KL Divergence value
    """
    # Tried with eps 1e-30 and this improves the precision by a small margin but overall ranking remains the same
    return target * torch.log((target / (output + eps)) + eps)


def calculate_frechet_distance(mu1, sigma1, mu2, sigma2, eps=1e-6):
    """Frechet Distance Implementation from https://github.com/bioinf-jku/TTUR/blob/master/fid.py.

    Numpy implementation of the Frechet Distance.
    The Frechet distance between two multivariate Gaussians X_1 ~ N(mu_1, C_1)
    and X_2 ~ N(mu_2, C_2) is
            d^2 = ||mu_1 - mu_2||^2 + Tr(C_1 + C_2 - 2*sqrt(C_1*C_2)).

    Stable version by Dougal J. Sutherland.

    Params:
    -- mu1   : Numpy array containing the activations of a layer of the
               inception net (like returned by the function 'get_predictions')
               for generated samples.
    -- mu2   : The sample mean over activations, precalculated on an
               representative data set.
    -- sigma1: The covariance matrix over activations for generated samples.
    -- sigma2: The covariance matrix over activations, precalculated on an
               representative data set.

    Raises:
        ValueError: Value error if imaginary component has large value.

    Returns:
    --   : The Frechet Distance.
    """
    mu1 = np.atleast_1d(mu1)
    mu2 = np.atleast_1d(mu2)

    sigma1 = np.atleast_2d(sigma1)
    sigma2 = np.atleast_2d(sigma2)

    assert (
        mu1.shape == mu2.shape
    ), "Training and test mean vectors have different lengths"
    assert (
        sigma1.shape == sigma2.shape
    ), "Training and test covariances have different dimensions"

    diff = mu1 - mu2

    # Product might be almost singular
    covmean = linalg.sqrtm(sigma1.dot(sigma2))
    if not np.isfinite(covmean).all():
        msg = (
            "fid calculation produces singular product; "
            "adding %s to diagonal of cov estimates"
        ) % eps
        print(msg)
        offset = np.eye(sigma1.shape[0]) * eps
        covmean = linalg.sqrtm((sigma1 + offset).dot(sigma2 + offset))

    # Numerical error might give slight imaginary component
    if np.iscomplexobj(covmean):
        if not np.allclose(np.diagonal(covmean).imag, 0, atol=1e-3):
            m = np.max(np.abs(covmean.imag))
            raise ValueError("Imaginary component {}".format(m))
        covmean = covmean.real

    tr_covmean = np.trace(covmean)

    return diff.dot(diff) + np.trace(sigma1) + np.trace(sigma2) - 2 * tr_covmean



"""Utilities file."""

from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

import torch as th
from PIL import Image


def _parse_args():
    """Argument parser."""
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "--batch-size",
        type=int,
        default=128,
        help="Batch size for wavelet packet transform.",
    )
    parser.add_argument(
        "--num-processes", type=int, default=None, help="Number of multiprocess."
    )
    parser.add_argument(
        "--save-packets", action="store_true", help="Save the packets as npz file."
    )
    parser.add_argument(
        "--wavelet", type=str, default="sym5", help="Choice of wavelet."
    )
    parser.add_argument(
        "--max_level", type=int, default=4, help="wavelet decomposition level"
    )
    parser.add_argument(
        "--log_scale", action="store_true", help="Use log scaling for wavelets."
    )
    parser.add_argument(
        "--deterministic",
        action="store_true",
        help="Set PyTorch to deterministic mode, for perfect reproducability.",
    )
    parser.add_argument(
        "path",
        type=str,
        nargs=2,
        help="Path to the generated images or path to .npz statistics file.",
    )
    return parser.parse_args()


class ImagePathDataset(th.utils.data.Dataset):
    """Image dataset."""

    def __init__(self, files, transforms=None):
        """File initialization."""
        self.files = files
        self.transforms = transforms

    def __len__(self):
        """Length of dataset."""
        return len(self.files)

    def __getitem__(self, i):
        """Load the image."""
        path = self.files[i]
        img = Image.open(path).convert("RGB")
        if self.transforms is not None:
            img = self.transforms(img)
        return img



"""Test wavelet packet transform."""

from itertools import product
from typing import Tuple

import numpy as np
import ptwt
import pytest
import pywt
import scipy
import torch as th

from pytorchfwd.freq_math import (
    forward_wavelet_packet_transform,
    generate_frequency_packet_image,
)


@pytest.mark.slow
def test_wpt():
    """Packet test for forward transfrom."""
    import scipy.datasets

    face = th.from_numpy(np.array(scipy.datasets.face()))
    face = th.stack([face, face, face, face], axis=0)
    face = face.type(th.FloatTensor) / 255.0
    b, h, w, c = face.shape
    face = th.reshape(face, (b, c, h, w))

    packets = forward_wavelet_packet_transform(
        face, wavelet="Haar", max_level=3, log_scale=False
    )
    p_image = generate_frequency_packet_image(packets, 3)
    assert p_image.shape == (4, 3, 768, 1024)


def test_inverse_wp():
    """Packet test for forward transfrom check with inverse transfrom."""
    face = th.Tensor(scipy.datasets.face())
    face = th.stack([face, face, face, face], axis=0)
    face = face.type(th.FloatTensor) / 255.0
    b, h, w, c = face.shape
    face = th.reshape(face, (b, c, h, w))

    packets = forward_wavelet_packet_transform(
        face, max_level=3, wavelet="db3", log_scale=False
    )

    reconstruction = _inverse_wavelet_packet_transform(
        packets, max_level=3, wavelet="db3"
    )
    assert th.max(th.abs(reconstruction[:, :, :768, :1024] - face)) < 1e-5


def _fold_channels(input_tensor: th.Tensor) -> th.Tensor:
    """Fold a trailing (color-) channel into the batch dimension.

    Args:
        input_tensor (th.Tensor): An array of shape [B, C, H, W]

    Returns:
        th.Tensor: The folded [B*C, H, W] image.
    """
    shape = input_tensor.shape
    return th.reshape(input_tensor, (-1, shape[-2], shape[-1]))


def _unfold_channels(
    input_tensor: th.Tensor, original_shape: Tuple[int, ...]
) -> th.Tensor:
    """Restore channels from the leading batch-dimension.

    Args:
        input_tensor (th.Tensor): An [B*C, packets, H, W] input array.
        original_shape (Tuple[int, ...]): Tuple containing shape of original input [BS, C, H, W].

    Returns:
        th.Tensor: Output of shape [B, packets, H, W, C]
    """
    _, packets, _, _ = input_tensor.shape
    b, c, h, w = original_shape
    return th.reshape(input_tensor, (b, packets, c, h, w))


def _inverse_wavelet_packet_transform(
    packet_tensor: th.Tensor, wavelet: str, max_level: int
):
    batch, _, channels, _, _ = packet_tensor.shape

    def get_node_order(level):
        wp_natural_path = list(product(["a", "h", "v", "d"], repeat=level))
        return ["".join(p) for p in wp_natural_path]

    wp_dict = {}
    for pos, path in enumerate(get_node_order(max_level)):
        wp_dict[path] = packet_tensor[:, pos, :, :, :]

    for level in reversed(range(max_level)):
        for node in get_node_order(level):
            data_a = _fold_channels(wp_dict[node + "a"])
            data_h = _fold_channels(wp_dict[node + "h"])
            data_v = _fold_channels(wp_dict[node + "v"])
            data_d = _fold_channels(wp_dict[node + "d"])
            rec = ptwt.waverec2(
                [data_a, (data_h, data_v, data_d)], pywt.Wavelet(wavelet)
            )
            height = rec.shape[1]
            width = rec.shape[2]
            rec = _unfold_channels(
                th.unsqueeze(rec, 1), [batch, channels, height, width]
            )
            rec = th.squeeze(rec, 1)
            wp_dict[node] = rec
    return rec



"""Test Wavelet packet Frechet distance."""

import os
from copy import deepcopy
from itertools import pairwise
from typing import Tuple

import numpy as np
import pytest
import torch as th
from sklearn.datasets import load_sample_images
from torchvision import transforms

from pytorchfwd.fwd import _compute_avg_frechet_distance, compute_packet_statistics

os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"

th.set_default_dtype(th.float64)
th.use_deterministic_algorithms(True)

make_dataloader = lambda x: th.utils.data.DataLoader(  # type: ignore # noqa: E731
    th.utils.data.TensorDataset(x), batch_size=1, shuffle=False, drop_last=False
)
default_params = {
    "dataloader": None,
    "wavelet": "Haar",
    "max_level": 1,
    "log_scale": False,
}


@pytest.mark.slow
def get_images(img_size: int = 32) -> th.Tensor:
    """Generate images of given size.

    Args:
        img_size (int): Integer specifying the desired image size. Defaults to 32.

    Returns:
        th.Tensor: Tensor containing images of shape [batchsize, channels, height, width].

    """
    dataset = load_sample_images()
    tower = th.Tensor(deepcopy(dataset.images[0]))
    flower = th.Tensor(deepcopy(dataset.images[1]))
    images = th.stack(
        [tower, tower, tower, tower, flower, flower, flower, flower], axis=0
    )
    images = images / 255.0
    images = th.permute(images, [0, 3, 1, 2])
    images = transforms.functional.resize(images, (img_size, img_size))
    assert images.shape == (8, 3, img_size, img_size)
    return images


def _calc_fwd(target_images: th.Tensor, output_images: th.Tensor) -> float:
    """Compute Frechet Wavelet Distance.

    Args:
        target_images (th.Tensor): Ground truth images.
        output_images (th.Tensor): Generated images.

    Returns:
        float: FWD value.
    """
    default_params["dataloader"] = make_dataloader(target_images)
    mu1, sigma1 = compute_packet_statistics(**default_params)
    default_params["dataloader"] = make_dataloader(output_images)
    mu2, sigma2 = compute_packet_statistics(**default_params)

    distance = _compute_avg_frechet_distance(mu1, mu2, sigma1, sigma2)
    return distance


def test_same_input():
    """FWD-test for same input."""
    target_images = get_images()
    output_images = deepcopy(target_images)
    fwd = _calc_fwd(target_images, output_images)
    assert np.allclose(fwd, 0.0, atol=1e-3)


def test_shuffle_input():
    """FWD-test for shuffled input."""
    target_images = get_images()
    # Shuffle the output images
    output_images = deepcopy(target_images)
    permutation = th.randperm(len(target_images))
    shuffled_images = output_images[permutation, :, :, :]
    assert not th.allclose(shuffled_images, output_images)

    shuffled_fwd = _calc_fwd(target_images, shuffled_images)
    unshuffled_fwd = _calc_fwd(target_images, output_images)
    assert np.allclose(shuffled_fwd, unshuffled_fwd, atol=1e-5)


@pytest.mark.slow
@pytest.mark.parametrize("wavelet", ["sym5", "db5", "Haar"])
@pytest.mark.parametrize("level", [1, 2])
def test_various_wavelets(wavelet: str, level: int):
    """FWD-test for various wavelets and transformation levels.

    Args:
        wavelet (str): Type of wavelet. Accepted values [sym5, db5, Haar].
        level (int): Packet transformation level. Accepted values [1, 2].
    """
    target_images = get_images()
    output_images = deepcopy(target_images)

    default_params["wavelet"] = wavelet
    default_params["max_level"] = level

    fwd = _calc_fwd(target_images, output_images)
    assert np.allclose(fwd, 0.0, atol=1e-3)


@pytest.mark.slow
@pytest.mark.parametrize("img_size_level", [(32, 1), (64, 2)])
def test_various_image_sizes(img_size_level: Tuple[int, int]):
    """FWD-test for various image sizes and transformation levels.

    Args:
        img_size_level (Tuple[int, int]): Tuple containing image size and transformation level.
    """
    size, level = img_size_level

    default_params["max_level"] = level

    target_images = get_images(size)
    output_images = deepcopy(target_images)
    assert output_images.shape == (8, 3, size, size)

    fwd = _calc_fwd(target_images, output_images)
    assert np.allclose(fwd, 0.0, atol=1e-3)


@pytest.mark.slow
def test_checkerboard_fwd():
    """FWD-Synthetic frequency test using various checkerboard sizes."""
    import scipy.ndimage

    range_max = 100
    # tile_size = 10 # determines tile size
    images = []

    for tile_size in [2, 10, 15]:
        # generate random grid

        grid = np.meshgrid(np.arange(0, range_max), np.arange(0, range_max))
        # grid = grid[0] + grid[1]

        tile_fun = lambda cord, ts: (cord % ts) < (ts // 2)  # type: ignore # noqa: E731

        x_tile = tile_fun(grid[0], tile_size)
        y_tile = tile_fun(grid[1], tile_size)
        grid = 1 - (x_tile + y_tile)

        grid = grid.astype(np.float32)

        # generate 10 images via a random fft-shift
        img_list = []
        for _ in range(1000):
            input_ = np.fft.fft2(grid)
            shift = np.random.uniform(-1, 1) * tile_size // 2
            image = scipy.ndimage.fourier_shift(input_, shift=shift)
            image = np.fft.ifft2(image)
            img_list.append(image.real)
        # B, H, W
        img_array = np.stack(img_list, axis=0)
        # B, C, H, W
        # img_array = np.expand_dims(img_array, 1)
        img_array = np.stack([img_array] * 3, 1)
        images.append(img_array)

    # a list for the wavelet frecet distances.
    fwd_list = []
    reference = images[0]
    default_params["max_level"] = 3
    for cimgs in images[1:]:
        fwd = _calc_fwd(
            target_images=th.from_numpy(reference), output_images=th.from_numpy(cimgs)
        )
        fwd_list.append(fwd)

    assert all(a < b for a, b in pairwise(fwd_list))
